/*
 * Copyright (c) University of Zagreb, Faculty of Electrical Engineering and Computing
 * Authors: Nikola Banic <nikola.banic@fer.hr> and Sven Loncaric <sven.loncaric@fer.hr>
 * 
 * This is only a research code and is therefore only of prototype quality.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * LITERATURE:
 * 
 * N. Banic and S. Loncaric
 * "Improving the White Patch method by sampling"
 * not published yet
 */

#include <cstdio>
#include <cmath>

#if defined(_WIN32) || defined(_WIN64)
#include <cv.h>
#include <highgui.h>
#else
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/highgui/highgui.hpp>
#endif

#define CV_LOAD_IMAGE_ANYDEPTH 2
#define CV_LOAD_IMAGE_ANYCOLOR 4

/**
	Estimates the global scene illumination.

	@param[in]	source Image with the scene for which the illumination is to be estimated for.
	@param[in]	n Random sample size.
	@param[in]	m Number of random samples.
	@return Normalized vector of the illumination estimation.
 */
cv::Scalar PerformIlluminationEstimation(cv::Mat img, int n, int m){
    cv::Mat source;
    img.convertTo(source, CV_64FC3);
   
    int rows=source.rows;
    int cols=source.cols;
   
    cv::Scalar result(0.0, 0.0, 0.0);
    cv::RNG random;
   
    for (int i=0;i<m;++i){
        cv::Scalar max(0.0, 0.0, 0.0);
        for (int j=0;j<n;++j){
            double p1=random.uniform(0.0, 1.0);
            double p2=random.uniform(0.0, 1.0);
            int row=(rows-1)*p1;
            int col=(cols-1)*p2;
            cv::Vec3d point=*(((cv::Vec3d *)source.data)+row*cols+col);
            for (int k=0;k<3;++k){
                if (max[k]<point[k]){
                    max[k]=point[k];
                }
            }
        }
        result+=max;
    }

    double c=result[0];
    result[0]=result[2];
    result[2]=c;
    double sum=result[0]*result[0]+result[1]*result[1]+result[2]*result[2];

    sum/=3;
    sum=sqrt(sum);

    result/=sum;
   
    return result;
}

/**
	Performs very simple chromatic adaptation of an image based on the illumination estimation.

	@param[in]	source The image to be chromatically adapted.
	@param[out]	destination The chromatically adapted image.
	@param[in]	illuminationEstimation Vector of the illumination estimation.
 */
void RemoveColorCast(cv::Mat source, cv::Mat &destination, cv::Scalar illuminationEstimation){
	
	int rows=source.rows;
	int cols=source.cols;

	cv::Mat converted;
	source.convertTo(converted, CV_64FC3);

	cv::Vec3d *input=(cv::Vec3d *)converted.data;
	for (int i=0;i<rows;++i){
		for (int j=0;j<cols;++j){
			cv::Vec3d &point=*(input+i*cols+j);
			for (int k=0;k<3;++k){
				point[k]/=illuminationEstimation[2-k];
			}
		}
	}

	converted.convertTo(destination, source.type());
	
}

int main(int argc, char **argv){

	if (argc<3){
		printf("Usage: %s input_file output_file [N [M]]\n", argv[0]);
		printf("\tN           - sample size\n");
		printf("\tM           - number of samples\n");
		return 0;
	}

	int N=50;
	int M=10;
	
	cv::Mat img=cv::imread(argv[1], CV_LOAD_IMAGE_ANYDEPTH|CV_LOAD_IMAGE_ANYCOLOR);
	
	if (img.rows*img.cols==0){
		return 0;
	}

	if (argc>3){
		sscanf(argv[3], "%d", &N);
		if (argc>4){
			sscanf(argv[4], "%d", &M);
		}
	}

	cv::Scalar illuminationEstimation=PerformIlluminationEstimation(img, N, M);

	cv::Mat result;
	RemoveColorCast(img, result, illuminationEstimation);

	imwrite(argv[2], result);

	return 0;
}
